testGeneFamilies.regression<-function(...){
  testGeneFamilies(method="poissonregression", ...)
}

testGeneFamilies.dircomp<-function(...){
  testGeneFamilies(method="poissongroups", ...)
}

testGeneFamilies<-function(data, method="Binomial", samples=c(1, 2), multtest="BH", ord=TRUE, 
                           design.matrix=NULL, coef=1, covariates=NULL, log.covariates=FALSE, groups=NULL,
			   overdisp=FALSE){
			     
  ncols<-dim(data$Data)[2]
  
  Statistics<-list()
  
  # Call the right method  
  if(tolower(method)=="binomial"){
    
    res.stat<-testGeneFamilies.binomial(data$Data[,samples])
    Statistics$RankingType<-"Binomial"
    Statistics$samples<-samples
    
  } else if(tolower(method)=="hypergeometric"){
  
    res.stat<-testGeneFamilies.hypergeometric(data$Data[,samples])    
    Statistics$RankingType<-"Hypergeometric"
    Statistics$samples<-samples
    
  } else if(tolower(method)=="gaussian"){
  
    res.stat<-testGeneFamilies.gaussian(data$Data[,samples])
    Statistics$RankingType<-"Gaussian"
    Statistics$samples<-samples
    
  } else if(tolower(method)=="poisson"){
  
    if(missing(design.matrix)){
      stop(sprintf("Error! Design and contrast matrix necessary for the general Poisson model.\n"))
    }       
    
    res.stat<-testGeneFamilies.poisson(data$Data, design.matrix, coef=coef, overdisp=overdisp)
    Statistics$RankingType<-"Poisson"
    Statistics$design.matrix<-design.matrix
    Statistics$coef<-coef
    
  } else if (tolower(method)=="poissonregression"){
  
    if(is.null(covariates)){    
      stop(sprintf("Error! Covariates necessary for Poisson regression model.\n"))
    }    

    Statistics$RankingType<-"PoissonRegression"
    Statistics$coef<-2
    
    if(log.covariates==TRUE){
      Statistics$covariates<-log10(covariates)
      Statistics$covariates.antilog<-covariates
    } else { 
      Statistics$covariates<-covariates
    }
    
    design.matrix<-cbind(rep(1, length(Statistics$covariates)), Statistics$covariates)
    Statistics$design.matrix<-design.matrix 
    res.stat<-testGeneFamilies.poisson(data$Data, Statistics$design.matrix, coef=Statistics$coef, overdisp=overdisp)
    
  } else if (tolower(method)=="poissongroups"){
  
    if(is.null(groups)){
      stop(sprintf("Error! Groups necessray for Poisson direct comparison.\n"))
    }
    
    if(length(unique(groups))!=2){
      stop(sprintf("Error! Only two groups allowed in Poisson direct comparison.\n"))
    }
    
    Statistics$RankingType<-"PoissonGroups"
    Statistics$coef<-2
    Statistics$groups<-groups        

    design.matrix<-cbind(rep(1, length(Statistics$groups)), as.integer(Statistics$groups==unique(Statistics$groups)[1]))
    Statistics$design.matrix<-design.matrix
    res.stat<-testGeneFamilies.poisson(data$Data, Statistics$design.matrix, coef=Statistics$coef, overdisp=overdisp)
    
  } else {  
    stop(sprintf("Unknown method: %s.\n", method))
  }
   
  # Save the details
  Statistics$Ranking.details<-res.stat$Ranking.details
  Statistics$Ranking.details$Annotation<-data$Annotation
   
  # Create a data frame with the results
  Statistics$Ranking<-data.frame(data$Annotation[,1], res.stat$PValues, data$Annotation[,2])
  colnames(Statistics$Ranking)<-c(colnames(data$Annotation)[1], colnames(res.stat$PValues), colnames(data$Annotation)[2])
 
  # Multple testing stuff.
  if(!is.null(multtest)){
    # Call multtest correction procedure
    multtest.res<-mt.rawp2adjp(Statistics$Ranking[,"P-value"], proc=multtest)
   
    # Return original order
    multtest.p<-multtest.res$adjp[order(multtest.res$index),2]
   
    # Save information regarding multiple testing procedue
    Statistics$Multtest<-multtest
   
    # Create a new data.frame
    Statistics$Ranking<-data.frame(data$Annotation[,1], res.stat$PValues, multtest.p, data$Annotation[,2])
    colnames(Statistics$Ranking)<-c(colnames(data$Annotation)[1], colnames(res.stat$PValues), sprintf("Adjusted P-value (%s)", multtest), colnames(data$Annotation)[2])
  }

  # Add data and annotation
  Statistics$Data<-data$Data
  Statistics$Annotation<-data$Annotation
  

  # Sort list
  if(ord==TRUE){
    ord.ranking<-order(Statistics$Ranking[,"P-value"])
    Statistics$Ranking<-Statistics$Ranking[ord.ranking,]
  }
  
  Statistics$Type<-data$Type

  return(Statistics)

}

testGeneFamilies.binomial<-function(data){

  n<-dim(data)[1]

  M.1<-sum(data[,1])
  M.2<-sum(data[,2])  

  para.1<-M.1/(M.1+M.2)
  para.2<-M.2/(M.1+M.2)
  
  N<-apply(data,1,sum)
  
  p.1<-pbinom(q=data[,1], size=N, p=para.1)
  p.2<-pbinom(q=data[,1]-1, size=N, p=para.1, lower.tail=FALSE)
    
  p.binom<-pmin.int(pmin.int(p.1, p.2)*2, rep(1,n))
  
  Result<-data.frame(data[,1], data[,2], N, p.binom)
  colnames(Result)<-c("Metagenome 1", "Metagenome 2", "Total Count", "P-value")
  
  # Additional details
  Details<-list()
  Details$TotalCounts<-c(M.1, M.2)  
  Details$data<-data
  
  return(list(PValues=Result, Ranking.details=Details))

}

testGeneFamilies.hypergeometric<-function(data){

  n<-dim(data)[1]

  M.1<-sum(data[,1])
  M.2<-sum(data[,2])  

  N<-apply(data,1,sum)
 
  p.hyper<-rep(NA, length(dim(data)[1]))

  for(i in 1:n){   
      mat<-matrix(c(data[i,1], data[i,2], M.1-data[i,1], M.2-data[i,2]), byrow=TRUE, nc=2, nr=2)
      p.hyper[i]<-fisher.test(mat, alternative="two.sided")$p.value
  }
  
  Result<-data.frame(data, N, p.hyper)
  colnames(Result)<-c("Sample 1", "Sample 2", "Total Count", "P-value")
 
  Details<-list()
  Details$TotalCounts<-c(M.1, M.2)
  Details$data<-data
 
  return(list(PValues=Result, Ranking.details=Details))
}

testGeneFamilies.gaussian<-function(data){

  M.1<-sum(data[,1])
  M.2<-sum(data[,2])  

  N<-apply(data,1,sum) 

  f1<-data[,1]/M.1
  f2<-data[,2]/M.2
  f<-(data[,1]+data[,2])/(M.1+M.2)
  deltamu<-f1-f2
  stderr<-sqrt(f*(1-f)*(1/M.1+1/M.2))
  stat.norm<-deltamu/stderr  
  stat.norm[is.nan(stat.norm)]<-0
  p.norm<-2*(1-pnorm(stat.norm))
  
  Result<-data.frame(stat.norm, N, p.norm)
  colnames(Result)<-c("Gaussian sum", "Total Count", "P-value")

  Details<-list()
  Details$TotalCounts<-c(M.1, M.2)
  Details$deltamu<-deltamu
  Details$stderr<-stderr
  Details$data<-data

  return(list(PValues=Result, Ranking.details=Details))
}

testGeneFamilies.poisson<-function(data, design.matrix, coef, overdisp=FALSE){

  n<-dim(data)[1]
  p.poisson<-numeric(length=n)
  est.coef<-numeric(length=n)
  R2<-numeric(length=n)
  AIC<-numeric(length=n)

  counts.total<-apply(data, 2, sum)

  # Loop over all COGs
  for(i in 1:n){
  
    res.po<-testGeneFamilies.poisson.fit(data[i,], counts.total, design.matrix, overdisp=overdisp)

    # Get p-value and value of the coefficient
    p.poisson[i]<-summary(res.po$fit)$coef[coef,4]
    est.coef[i]<-summary(res.po$fit)$coef[coef,1]
    R2[i]<-res.po$R2
    AIC[i]<-res.po$AIC
  }
    
  Result<-data.frame(est.coef, AIC, p.poisson)
  colnames(Result)<-c("Coefficient", "AIC", "P-value")
  
  # Add R2, AIC, stderr, etc
  Details<-list()
  Details$TotalCounts<-counts.total
  Details$design.matrix<-design.matrix
  Details$R2<-R2
  Details$data<-data
  Details$AIC<-AIC
  
  return(list(PValues=Result, Ranking.details=Details))

}

testGeneFamilies.poisson.fit<-function(data.vec, counts.total, design.matrix, overdisp=FALSE){
  
  n.coef<-dim(design.matrix)[2]
  data.mat<-data.frame(t(data.vec), counts.total, design.matrix)
  colnames(data.mat)<-c("Counts", "Total", paste("Coef", 1:n.coef, sep=""))
  
  if(overdisp==FALSE){
    family.glm<-poisson
  } else { 
    family.glm<-quasipoisson
  }
  
  form<-paste("Counts ~ 0 + offset(log(Total))+", paste(paste("Coef", 1:n.coef, sep=""), collapse="+"), sep="")
  
  res.glm<-glm(formula(form), data=data.mat, family=family.glm)
    
  Result<-list()
  
  Result$fit<-res.glm
  Result$n.coef<-n.coef
  Result$data.mat<-data.mat
  Result$form<-form
  Result$R2<-calculatePoissonR2(res.glm, k=n.coef-1)
  Result$AIC<-res.glm$aic
 
  return(Result)

}

calculatePoissonR2<-function(res.poiss, k=1){

  y<-res.poiss$y
  muhat<-res.poiss$fitted.values
  ymean<-mean(res.poiss$y)
 
  ylogy<-y*log(y)
  ylogy[is.nan(ylogy)]<-0


  # log(y) when y is zero?? check ref!!!!!
  S1<-sum((ylogy-y)-(y*log(muhat)-muhat))
  S2<-sum((ylogy-y)-(y*log(ymean)-ymean))

  R2<-(1-S1/S2)

  return(R2)
}
